@@ -1,4 +1,4 @@
-/*   
+/*
 
   Copyright 2004-2012, Martian Software, Inc.
 
@@ -13,7 +13,7 @@
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
-  
+
 */
 
 /**
@@ -29,6 +29,8 @@
 	#include <netdb.h>
 	#include <netinet/in.h>
 	#include <sys/socket.h>
+	#include <sys/un.h>
+	#include <sys/time.h>
 	#include <sys/types.h>
 #endif
 
@@ -37,6 +39,7 @@
 #include <string.h>
 #include <unistd.h>
 #include <fcntl.h>
+#include <errno.h>
 
 #define NAILGUN_VERSION "0.9.0"
 
@@ -55,8 +58,12 @@
 	#define FILE_SEPARATOR '/'
 	typedef int HANDLE;
 	typedef unsigned int SOCKET;
-	/* buffer used for reading an writing chunk data */
-	char buf[BUFSIZE];
+#endif
+
+#ifdef __APPLE__
+  #define SEND_FLAGS 0
+#else
+  #define SEND_FLAGS MSG_NOSIGNAL
 #endif
 
 #ifndef MIN
@@ -94,7 +101,12 @@
 #define CHUNKTYPE_DIR 'D'
 #define CHUNKTYPE_CMD 'C'
 #define CHUNKTYPE_EXIT 'X'
-#define CHUNKTYPE_STARTINPUT 'S'
+#define CHUNKTYPE_SENDINPUT 'S'
+#define CHUNKTYPE_HEARTBEAT 'H'
+
+#define HEARTBEAT_TIMEOUT_MILLIS 500
+
+#define NAILGUN_TTY_FORMAT "NAILGUN_TTY_%d=%d"
 
 /*
    the following is required to compile for hp-ux
@@ -110,8 +122,14 @@ int nailgunsocket = 0;
 /* buffer used for receiving and writing nail output chunks */
 char buf[BUFSIZE];
 
-/* track whether or not we've been told to send stdin to server */
-int startedInput = 0;
+/* track whether server is ready to receive */
+#ifdef WIN32
+    HANDLE readyToSend = 0;
+    HANDLE sending = 0;
+#else
+    int readyToSend = 0;
+    struct timeval sendtime;
+#endif
 
 /**
  * Clean up the application.
@@ -141,9 +159,9 @@ void handleError () {
   LPVOID lpMsgBuf;
   int error = GetLastError();
 
-  FormatMessage( 
-    FORMAT_MESSAGE_ALLOCATE_BUFFER | 
-    FORMAT_MESSAGE_FROM_SYSTEM | 
+  FormatMessage(
+    FORMAT_MESSAGE_ALLOCATE_BUFFER |
+    FORMAT_MESSAGE_FROM_SYSTEM |
     FORMAT_MESSAGE_IGNORE_INSERTS,
     NULL,
     error,
@@ -157,11 +175,18 @@ void handleError () {
 
   /* Free the buffer. */
   LocalFree( lpMsgBuf );
- 
+
   cleanUpAndExit(error);
 }
 #endif
 
+/**
+ * Exits the client if the nailgun server ungracefully shut down the connection.
+ */
+void handleSocketClose() {
+  cleanUpAndExit(NAILGUN_CONNECTION_BROKEN);
+}
+
 /**
  * Writes everything in the specified buffer to the specified
  * socket handle.
@@ -173,33 +198,35 @@ void handleError () {
  * @return total bytes written or 0 if failure
  */
 int sendAll(SOCKET s, char *buf, int len) {
-  int total = 0;      
-  int bytesleft = len; 
+  int total = 0;
+  int bytesleft = len;
   int n = 0;
-    
+
   while(total < len) {
-    n = send(s, buf+total, bytesleft, 0);
-    
-    if (n == -1) { 
+    n = send(s, buf+total, bytesleft, SEND_FLAGS);
+
+    if (n == -1) {
       break;
     }
-    
+
     total += n;
     bytesleft -= n;
   }
 
-  return n==-1 ? 0:total; 
+  return n==-1 ? 0:total;
 }
 
 /**
- * Sends a chunk header noting the specified payload size and chunk type.
- * 
+ * Sends a chunk noting the specified payload size and chunk type.
+ * Waits for sending mutex on Win32.
+ *
  * @param size the payload size
  * @param chunkType the chunk type identifier
  */
-void sendHeader(unsigned int size, char chunkType) {
+void sendChunk(unsigned int size, char chunkType, char* buf) {
   /* buffer used for reading and writing chunk headers */
   char header[CHUNK_HEADER_LEN];
+  int bytesSent;
 
   header[0] = (size >> 24) & 0xff;
   header[1] = (size >> 16) & 0xff;
@@ -207,7 +234,25 @@ void sendHeader(unsigned int size, char chunkType) {
   header[3] = size & 0xff;
   header[4] = chunkType;
 
-  sendAll(nailgunsocket, header, CHUNK_HEADER_LEN);
+#ifdef WIN32
+  if (WaitForSingleObject(sending, INFINITE) != WAIT_OBJECT_0) {
+    handleError();
+  }
+#else
+  gettimeofday(&sendtime, NULL);
+#endif
+
+  bytesSent = sendAll(nailgunsocket, header, CHUNK_HEADER_LEN);
+  if (bytesSent != 0 && size > 0) {
+    bytesSent = sendAll(nailgunsocket, buf, size);
+  } else if (bytesSent == 0 && (chunkType != CHUNKTYPE_HEARTBEAT || !(errno == EPIPE || errno == ECONNRESET))) {
+    perror("send");
+    handleSocketClose();
+  }
+
+#ifdef WIN32
+  ReleaseMutex(sending);
+#endif
 }
 
 /**
@@ -220,24 +265,23 @@ void sendHeader(unsigned int size, char chunkType) {
  */
 int sendFileArg(char *filename) {
   int i, f;
-  
+
   if ((f = open(filename, O_RDONLY)) < 0) {
     perror("--nailgun-filearg");
-    return 1;  
+    return 1;
   }
 
   i = read(f, buf, BUFSIZE);
   while (i > 0) {
-    sendHeader(i, CHUNKTYPE_LONGARG);
-    sendAll(nailgunsocket, buf, i);
+    sendChunk(i, CHUNKTYPE_LONGARG, buf);
     i = read(f, buf, BUFSIZE);
   }
   if (i < 0) {
     perror("--nailgun-filearg");
     return 1;
   }
-  sendHeader(0, CHUNKTYPE_LONGARG);
-  
+  sendChunk(0, CHUNKTYPE_LONGARG, buf);
+
   close(f);
   return 0;
 }
@@ -250,15 +294,7 @@ int sendFileArg(char *filename) {
  */
 void sendText(char chunkType, char *text) {
   int len = text ? strlen(text) : 0;
-  sendHeader(len, chunkType);
-  sendAll(nailgunsocket, text, len);
-}
-
-/**
- * Exits the client if the nailgun server ungracefully shut down the connection.
- */
-void handleSocketClose() {
-  cleanUpAndExit(NAILGUN_CONNECTION_BROKEN);
+  sendChunk(len, chunkType, text);
 }
 
 /**
@@ -271,16 +307,17 @@ void handleSocketClose() {
 void recvToFD(HANDLE destFD, char *buf, unsigned long len) {
   unsigned long bytesRead = 0;
   int bytesCopied;
-  
+
   while (bytesRead < len) {
     unsigned long bytesRemaining = len - bytesRead;
     int bytesToRead = (BUFSIZE < bytesRemaining) ? BUFSIZE : bytesRemaining;
     int thisPass = 0;
-    
+
     thisPass = recv(nailgunsocket, buf, bytesToRead, MSG_WAITALL);
-    if (thisPass < bytesToRead) handleSocketClose();
-    
-   
+    if (thisPass == 0 || thisPass == -1) {
+      perror("recv");
+      handleSocketClose();
+    }
     bytesRead += thisPass;
 
     bytesCopied = 0;
@@ -298,12 +335,29 @@ void recvToFD(HANDLE destFD, char *buf, unsigned long len) {
 
         bytesCopied += thisWrite;
       #else
-        bytesCopied += write(destFD, buf + bytesCopied, thisPass - bytesCopied);
+        int bytesWritten = write(destFD, buf + bytesCopied, thisPass - bytesCopied);
+        if (bytesWritten == -1) {
+          perror("write");
+          handleSocketClose();
+        }
+        bytesCopied += bytesWritten;
       #endif
-    }  
+    }
   }
 }
 
+unsigned long recvToBuffer(unsigned long len) {
+  unsigned long bytesRead = 0;
+  while(bytesRead < len) {
+    int thisPass = recv(nailgunsocket, buf + bytesRead, len - bytesRead, MSG_WAITALL);
+    if (thisPass == 0 || thisPass == -1) {
+        perror("recv");
+        handleSocketClose();
+    }
+    bytesRead += thisPass;
+  }
+  return bytesRead;
+}
 
 /**
  * Processes an exit chunk from the server.  This is just a string
@@ -315,16 +369,16 @@ void recvToFD(HANDLE destFD, char *buf, unsigned long len) {
 void processExit(char *buf, unsigned long len) {
   int exitcode;
   int bytesToRead = (BUFSIZE - 1 < len) ? BUFSIZE - 1 : len;
-  int bytesRead = recv(nailgunsocket, buf, bytesToRead, MSG_WAITALL);
-  
+  int bytesRead = recvToBuffer(bytesToRead);
+
   if (bytesRead < 0) {
     handleSocketClose();
   }
-  
+
   buf[bytesRead] = 0;
-  
+
   exitcode = atoi(buf);
-  
+
   cleanUpAndExit(exitcode);
 }
 
@@ -336,19 +390,43 @@ void processExit(char *buf, unsigned long len) {
  * @param len the number of bytes to send
  */
 void sendStdin(char *buf, unsigned int len) {
-  sendHeader(len, CHUNKTYPE_STDIN);
-  sendAll(nailgunsocket, buf, len);
+#ifndef WIN32
+  readyToSend = 0;
+#endif
+  sendChunk(len, CHUNKTYPE_STDIN, buf);
 }
 
 /**
  * Sends a stdin-eof chunk to the nailgun server
  */
 void processEof() {
-  sendHeader(0, CHUNKTYPE_STDIN_EOF);
+  sendChunk(0, CHUNKTYPE_STDIN_EOF, buf);
 }
 
+/**
+ * Sends a heartbeat chunk to let the server know the client is still alive.
+ */
+void sendHeartbeat() {
+  sendChunk(0, CHUNKTYPE_HEARTBEAT, buf);
+}
 
 #ifdef WIN32
+
+HANDLE createEvent(BOOL manualReset) {
+  return CreateEvent(NULL, /* default security */
+		     manualReset,
+		     FALSE, /* initial state unsignalled */
+		     NULL /* unnamed event */);
+}
+
+DWORD WINAPI sendHeartbeats(LPVOID lpParameter) {
+	/* this could be made more efficient by only sending heartbeats when stdin chunks aren't being sent */
+	for (;;) {
+		Sleep(HEARTBEAT_TIMEOUT_MILLIS);
+		sendHeartbeat();
+	}
+}
+
 /**
  * Thread main for reading from stdin and sending
  */
@@ -356,23 +434,32 @@ DWORD WINAPI processStdin (LPVOID lpParameter) {
   /* buffer used for reading and sending stdin chunks */
   char wbuf[BUFSIZE];
 
-  for (;;) {
-    DWORD numberOfBytes = 0;
+  /* number of bytes read */
+  DWORD numberOfBytes;
 
-    if (!ReadFile(NG_STDIN_FILENO, wbuf, BUFSIZE, &numberOfBytes, NULL)) {
-      if (numberOfBytes != 0) {
-        handleError();
-      }
+  for (;;) {
+
+    /* wait for ready to send */
+    if(WaitForSingleObject(readyToSend, INFINITE) != WAIT_OBJECT_0) {
+      handleError();
     }
 
+	/* read data from stdin */
+	if (! ReadFile(NG_STDIN_FILENO, wbuf, BUFSIZE, &numberOfBytes, NULL)) {
+		if (numberOfBytes != 0) {
+			handleError();
+		}
+	}
+
+    /* send data to server */
     if (numberOfBytes > 0) {
       sendStdin(wbuf, numberOfBytes);
     } else {
       processEof();
       break;
     }
-  }
 
+  }
   return 0;
 }
 #else
@@ -399,8 +486,12 @@ int processStdin() {
  */
 void initSockets () {
   WSADATA win_socket_data;     /* required to initialise winsock */
-  
+
   WSAStartup(2, &win_socket_data);
+
+  /* create flow control event and mutex */
+  readyToSend = createEvent(FALSE);
+  sending = CreateMutex(NULL, FALSE, NULL);
 }
 #endif
 
@@ -411,7 +502,7 @@ void initSockets () {
 void initIo () {
   /* create non-blocking console io */
   AllocConsole();
-  
+
   NG_STDIN_FILENO = GetStdHandle(STD_INPUT_HANDLE);
   NG_STDOUT_FILENO = GetStdHandle(STD_OUTPUT_HANDLE);
   NG_STDERR_FILENO = GetStdHandle(STD_ERROR_HANDLE);
@@ -429,10 +520,14 @@ void winStartInput () {
   securityAttributes.bInheritHandle = TRUE;
   securityAttributes.lpSecurityDescriptor = NULL;
   securityAttributes.nLength = 0;
-  
+
   if (!CreateThread(&securityAttributes, 0, &processStdin, NULL, 0, &threadId)) {
     handleError();
   }
+
+  if (!CreateThread(&securityAttributes, 0, &sendHeartbeats, NULL, 0, &threadId)) {
+    handleError();
+  }
 }
 #endif
 
@@ -442,23 +537,18 @@ void winStartInput () {
 void processnailgunstream() {
 
   /*for (;;) {*/
-    int bytesRead = 0;
     unsigned long len;
     char chunkType;
 
-    bytesRead = recv(nailgunsocket, buf, CHUNK_HEADER_LEN, MSG_WAITALL);
+    recvToBuffer(CHUNK_HEADER_LEN);
 
-    if (bytesRead < CHUNK_HEADER_LEN) {
-      handleSocketClose();
-    }
-  
     len = ((buf[0] << 24) & 0xff000000)
       | ((buf[1] << 16) & 0x00ff0000)
       | ((buf[2] << 8) & 0x0000ff00)
       | ((buf[3]) & 0x000000ff);
-  
+
     chunkType = buf[4];
-  
+
     switch(chunkType) {
       case CHUNKTYPE_STDOUT: recvToFD(NG_STDOUT_FILENO, buf, len);
             break;
@@ -466,13 +556,12 @@ void processnailgunstream() {
             break;
       case CHUNKTYPE_EXIT:   processExit(buf, len);
             break;
-      case CHUNKTYPE_STARTINPUT:
-            if (!startedInput) {
-                #ifdef WIN32
-                winStartInput();
-                #endif
-      		startedInput = 1;
-      	    }
+      case CHUNKTYPE_SENDINPUT:
+#ifdef WIN32
+            SetEvent(readyToSend);
+#else
+            readyToSend = 1;
+#endif
             break;
       default:  fprintf(stderr, "Unexpected chunk type %d ('%c')\n", chunkType, chunkType);
           cleanUpAndExit(NAILGUN_UNEXPECTED_CHUNKTYPE);
@@ -480,6 +569,17 @@ void processnailgunstream() {
   /*}*/
 }
 
+/**
+ * Returns the time interval between start and end in milliseconds.
+ * @param end the end time
+ * @param start the start time
+ */
+int intervalMillis(struct timeval end, struct timeval start) {
+
+  return ((end.tv_sec - start.tv_sec) * 1000) +
+    ((end.tv_usec - start.tv_usec) /1000);
+}
+
 /**
  * Trims any path info from the beginning of argv[0] to determine
  * the name used to launch the client.
@@ -498,12 +598,23 @@ char *shortClientName(char *s) {
  * @param s the program name to check
  */
 int isNailgunClientName(char *s) {
+  /* VMS can't get the command name from argv[0] when defined
+     as a foreign command, as the symbol is expanded. Aliases
+     are supported by defining a foreign command that includes
+     the class alias argument, e.g.
+         ng=="$path:[to]ng.exe"
+         alias==ng+" ""alias"""
+  */
+  #ifdef __VMS
+  return (1);
+  #else
   #ifdef WIN32
   return (!strcasecmp(s, NAILGUN_CLIENT_NAME) ||
            !strcasecmp(s, NAILGUN_CLIENT_NAME_EXE));
   #else
   return(!(strcmp(s, NAILGUN_CLIENT_NAME)));
   #endif
+  #endif
 }
 
 /**
@@ -519,7 +630,7 @@ void usage(int exitcode) {
   fprintf(stderr, "          (to execute an aliased class, where \"alias\"\n");
   fprintf(stderr, "           is both the alias for the class and a symbolic\n");
   fprintf(stderr, "           link to the ng client)\n\n");
-  
+
   fprintf(stderr, "where options include:\n");
   fprintf(stderr, "   --nailgun-D<name>=<value>   set/override a client environment variable\n");
   fprintf(stderr, "   --nailgun-version           print product version and exit\n");
@@ -529,7 +640,7 @@ void usage(int exitcode) {
   fprintf(stderr, "                               if set, otherwise localhost)\n");
   fprintf(stderr, "   --nailgun-port              to specify the port of the nailgun server\n");
   fprintf(stderr, "                               (default is NAILGUN_PORT environment variable\n");
-  fprintf(stderr, "                               if set, otherwise 2113)\n");  
+  fprintf(stderr, "                               if set, otherwise 2113)\n");
   fprintf(stderr, "   --nailgun-filearg FILE      places the entire contents of FILE into the\n");
   fprintf(stderr, "                               next argument, which is interpreted as a string\n");
   fprintf(stderr, "                               using the server's default character set.  May be\n");
@@ -541,7 +652,12 @@ void usage(int exitcode) {
 
 int main(int argc, char *argv[], char *env[]) {
   int i;
-  struct sockaddr_in server_addr;
+  struct sockaddr *server_addr;
+  socklen_t server_addr_len;
+  struct sockaddr_in server_addr_in;
+  #ifndef WIN32
+    struct sockaddr_un server_addr_un;
+  #endif
   char *nailgun_server;        /* server as specified by user */
   char *nailgun_port;          /* port as specified by user */
   char *cwd;
@@ -549,28 +665,32 @@ int main(int argc, char *argv[], char *env[]) {
   struct hostent *hostinfo;
   char *cmd;
   int firstArgIndex;           /* the first argument _to pass to the server_ */
+  char isattybuf[] = NAILGUN_TTY_FORMAT;
 
   #ifndef WIN32
     fd_set readfds;
     int eof = 0;
+    struct timeval readtimeout;
+    struct timeval currenttime;
+    memset(&sendtime, '\0', sizeof(sendtime));
   #endif
 
   #ifdef WIN32
   initSockets();
   #endif
-  
+
   /* start with environment variable.  default to localhost if not defined. */
   nailgun_server = getenv("NAILGUN_SERVER");
   if (nailgun_server == NULL) {
     nailgun_server = "127.0.0.1";
   }
-  
+
   /* start with environment variable.  default to normal nailgun port if not defined */
   nailgun_port = getenv("NAILGUN_PORT");
   if (nailgun_port == NULL) {
     nailgun_port = NAILGUN_PORT_DEFAULT;
   }
-  
+
   /* look at the command used to launch this program.  if it was "ng", then the actual
      command to issue to the server must be specified as another argument.  if it
      wasn't ng, assume that the desired command name was symlinked to ng in the user's
@@ -585,14 +705,14 @@ int main(int argc, char *argv[], char *env[]) {
      display usage and exit.  Don't handle -h|--help if a command other than
      ng or ng.exe was used, since the appropriate nail should then handle
      --help. */
-  if (cmd == NULL && 
-        (argc == 1 || 
+  if (cmd == NULL &&
+        (argc == 1 ||
 	  (argc == 2 && strcmp("--help", argv[1]) == 0) ||
 	  (argc == 2 && strcmp("-h", argv[1]) == 0))) usage(0);
-     
+
   firstArgIndex = 1;
 
-  /* quite possibly the lamest commandline parsing ever. 
+  /* quite possibly the lamest commandline parsing ever.
      look for the two args we care about (--nailgun-server and
      --nailgun-port) and NULL them and their parameters after
      reading them if found.  later, when we send args to the
@@ -630,34 +750,64 @@ int main(int argc, char *argv[], char *env[]) {
   if (cmd == NULL) {
     usage(NAILGUN_BAD_ARGUMENTS);
   }
-  
-  /* jump through a series of connection hoops */  
-  hostinfo = gethostbyname(nailgun_server);
 
-  if (hostinfo == NULL) {
-    fprintf(stderr, "Unknown host: %s\n", nailgun_server);
-    cleanUpAndExit(NAILGUN_CONNECT_FAILED);
-  }
- 
-  port = atoi(nailgun_port);
+  #ifndef WIN32
+    if (strncmp(nailgun_server, "local:", 6) == 0) {
+      const char *socket_path = nailgun_server + 6;
+      size_t socket_path_len = strlen(socket_path);
+      if (socket_path_len > sizeof(server_addr_un.sun_path) - 1) {
+        fprintf(stderr, "Socket path [%s] too long (%ld)\n", socket_path, (long) socket_path_len);
+        cleanUpAndExit(NAILGUN_SOCKET_FAILED);
+      }
+      if ((nailgunsocket = socket(PF_LOCAL, SOCK_STREAM, 0)) == -1) {
+        perror("socket");
+        cleanUpAndExit(NAILGUN_SOCKET_FAILED);
+      }
 
-  if ((nailgunsocket = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
-    perror("socket");
-    cleanUpAndExit(NAILGUN_SOCKET_FAILED);
-  }
+      server_addr_un.sun_family = AF_LOCAL;
+      strncpy(server_addr_un.sun_path, socket_path, socket_path_len);
+      server_addr_un.sun_path[socket_path_len] = '\0';
 
-  server_addr.sin_family = AF_INET;    
-  server_addr.sin_port = htons(port);
-  server_addr.sin_addr = *(struct in_addr *) hostinfo->h_addr;
-  
-  memset(&(server_addr.sin_zero), '\0', 8);
+      #ifdef BSD
+        server_addr_un.sun_len = offsetof(struct sockaddr_un, sun_path) + socket_path_len;
+      #endif
+      server_addr = (struct sockaddr *)&server_addr_un;
+      server_addr_len = sizeof(server_addr_un);
+    } else {
+  #endif
 
-  if (connect(nailgunsocket, (struct sockaddr *)&server_addr,
-    sizeof(struct sockaddr)) == -1) {
+      /* jump through a series of connection hoops */
+      hostinfo = gethostbyname(nailgun_server);
+
+      if (hostinfo == NULL) {
+        fprintf(stderr, "Unknown host: %s\n", nailgun_server);
+        cleanUpAndExit(NAILGUN_CONNECT_FAILED);
+      }
+
+      port = atoi(nailgun_port);
+
+      if ((nailgunsocket = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
+        perror("socket");
+        cleanUpAndExit(NAILGUN_SOCKET_FAILED);
+      }
+
+      server_addr_in.sin_family = AF_INET;
+      server_addr_in.sin_port = htons(port);
+      server_addr_in.sin_addr = *(struct in_addr *) hostinfo->h_addr;
+
+      memset(&(server_addr_in.sin_zero), '\0', 8);
+      server_addr = (struct sockaddr *)&server_addr_in;
+      server_addr_len = sizeof(server_addr_in);
+
+  #ifndef WIN32
+    }
+  #endif
+
+  if (connect(nailgunsocket, server_addr, server_addr_len) == -1) {
     perror("connect");
     cleanUpAndExit(NAILGUN_CONNECT_FAILED);
-  } 
-    
+  }
+
   /* ok, now we're connected.  first send all of the command line
      arguments for the server, if any.  remember that we may have
      marked some arguments NULL if we read them to specify the
@@ -665,60 +815,84 @@ int main(int argc, char *argv[], char *env[]) {
   for(i = firstArgIndex; i < argc; ++i) {
     if (argv[i] != NULL) {
       if (!strcmp("--nailgun-filearg", argv[i])) {
-        sendFileArg(argv[++i]);
+        int sendResult = sendFileArg(argv[++i]);
+        if (sendResult != 0) {
+          perror("send");
+          handleSocketClose();
+        }
       } else sendText(CHUNKTYPE_ARG, argv[i]);
     }
   }
 
-  /* now send environment */  
+  /* now send environment */
   sendText(CHUNKTYPE_ENV, NAILGUN_FILESEPARATOR);
   sendText(CHUNKTYPE_ENV, NAILGUN_PATHSEPARATOR);
+#ifndef WIN32
+  /* notify isatty for standard pipes */
+  for(i = 0; i < 3; i++) {
+    sprintf(isattybuf, NAILGUN_TTY_FORMAT, i, isatty(i));
+    sendText(CHUNKTYPE_ENV, isattybuf);
+  }
+#endif
+  /* forward the client process environment */
   for(i = 0; env[i]; ++i) {
     sendText(CHUNKTYPE_ENV, env[i]);
   }
-  
+
   /* now send the working directory */
   cwd = getcwd(NULL, 0);
   sendText(CHUNKTYPE_DIR, cwd);
   free(cwd);
-  
+
   /* and finally send the command.  this marks the point at which
      streams are linked between client and server. */
   sendText(CHUNKTYPE_CMD, cmd);
 
-  
-  /* initialise the std-* handles and the thread to send stdin to the server */ 
+
+  /* initialise the std-* handles and the thread to send stdin to the server */
   #ifdef WIN32
   initIo();
+  winStartInput();
   #endif
 
-  /* stream forwarding loop */	
+  /* stream forwarding loop */
   while(1) {
     #ifndef WIN32
       FD_ZERO(&readfds);
 
       /* don't select on stdin if we've already reached its end */
-      if (startedInput && !eof) {
+      if (readyToSend && !eof) {
 	FD_SET(NG_STDIN_FILENO, &readfds);
       }
 
       FD_SET(nailgunsocket, &readfds);
-      if (select (nailgunsocket + 1, &readfds, NULL, NULL, NULL) == -1) {
-	perror("select");
+
+      memset(&readtimeout, '\0', sizeof(readtimeout));
+      readtimeout.tv_usec = HEARTBEAT_TIMEOUT_MILLIS * 1000;
+      if(select (nailgunsocket + 1, &readfds, NULL, NULL, &readtimeout) == -1) {
+	  perror("select");
       }
-	  
+
       if (FD_ISSET(nailgunsocket, &readfds)) {
     #endif
 	processnailgunstream();
     #ifndef WIN32
       } else if (FD_ISSET(NG_STDIN_FILENO, &readfds)) {
-	if (!processStdin()) {
-	  FD_CLR(NG_STDIN_FILENO, &readfds);
-	  eof = 1;
-	}
+        int result = processStdin();
+        if (result == -1) {
+          perror("read");
+          handleSocketClose();
+        } else if (result == 0) {
+          FD_CLR(NG_STDIN_FILENO, &readfds);
+          eof = 1;
+        }
+      }
+      gettimeofday(&currenttime, NULL);
+      if (intervalMillis(currenttime, sendtime) > HEARTBEAT_TIMEOUT_MILLIS) {
+	  sendHeartbeat();
       }
     #endif
-  }  
+  }
 
   /* normal termination is triggered by the server, and so occurs in processExit(), above */
 }
