@@ -26,9 +26,10 @@ import java.net.Socket;
 import java.net.UnknownHostException;
 import java.util.Iterator;
 import java.util.Map;
+import java.util.concurrent.atomic.AtomicBoolean;
 
 import com.martiansoftware.nailgun.builtins.DefaultNail;
-import java.util.Properties;
+import com.sun.jna.Platform;
 
 /**
  * <p>Listens for new connections from NailGun clients and launches NGSession
@@ -48,14 +49,9 @@ public class NGServer implements Runnable {
     public static final int DEFAULT_SESSIONPOOLSIZE = 10;
     
     /**
-     * The address on which to listen, or null to listen on all local addresses
+     * The address on which to listen
      */
-    private InetAddress addr = null;
-    
-    /**
-     * The port on which to listen, or zero to select a port automatically
-     */
-    private int port = 0;
+    private final NGListeningAddress listeningAddress;
     
     /**
      * The socket doing the listening
@@ -65,17 +61,17 @@ public class NGServer implements Runnable {
     /**
      * True if this NGServer has received instructions to shut down
      */
-    private boolean shutdown = false;
+    private final AtomicBoolean shutdown = new AtomicBoolean(false);
     
     /**
      * True if this NGServer has been started and is accepting connections
      */
-    private boolean running = false;
+    private final AtomicBoolean running = new AtomicBoolean(false);
     
     /**
      * This NGServer's AliasManager, which maps aliases to classes
      */
-    private AliasManager aliasManager;
+    private final AliasManager aliasManager;
     
     /**
      * If true, fully-qualified classnames are valid commands
@@ -91,7 +87,7 @@ public class NGServer implements Runnable {
     /**
      * A pool of NGSessions ready to handle client connections
      */
-    private NGSessionPool sessionPool = null;
+    private final NGSessionPool sessionPool;
     
     /**
      * <code>System.out</code> at the time of the NGServer's creation
@@ -111,12 +107,13 @@ public class NGServer implements Runnable {
     /**
      * a collection of all classes executed by this server so far
      */
-    private Map allNailStats = null;
+    private final Map allNailStats;
     
     /**
      * Remember the security manager we start with so we can restore it later
      */
     private SecurityManager originalSecurityManager = null;
+    private final int heartbeatTimeoutMillis;
 
     /**
      * Creates a new NGServer that will listen at the specified address and on
@@ -131,8 +128,8 @@ public class NGServer implements Runnable {
      * @param sessionPoolSize the max number of idle sessions allowed by the
      * pool
      */
-    public NGServer(InetAddress addr, int port, int sessionPoolSize) {
-        init(addr, port, sessionPoolSize);
+    public NGServer(InetAddress addr, int port, int sessionPoolSize, int timeoutMillis) {
+        this(new NGListeningAddress(addr, port), sessionPoolSize, timeoutMillis);
     }
 
     /**
@@ -145,11 +142,9 @@ public class NGServer implements Runnable {
      * @param addr the address at which to listen, or
      * <code>null</code> to bind to all local addresses
      * @param port the port on which to listen.
-     * @param sessionPoolSize the max number of idle sessions allowed by the
-     * pool
      */
     public NGServer(InetAddress addr, int port) {
-        init(addr, port, DEFAULT_SESSIONPOOLSIZE);
+        this(new NGListeningAddress(addr, port), DEFAULT_SESSIONPOOLSIZE, NGConstants.HEARTBEAT_TIMEOUT_MILLIS);
     }
 
     /**
@@ -160,26 +155,31 @@ public class NGServer implements Runnable {
      * <code>NGServer</code> and start it.
      */
     public NGServer() {
-        init(null, NGConstants.DEFAULT_PORT, DEFAULT_SESSIONPOOLSIZE);
+        this(new NGListeningAddress(null, NGConstants.DEFAULT_PORT), DEFAULT_SESSIONPOOLSIZE, NGConstants.HEARTBEAT_TIMEOUT_MILLIS);
     }
-    
+
     /**
-     * Sets up the NGServer internals
+     * Creates a new NGServer that will listen at the specified address and on
+     * the specified port with the specified session pool size. This does
+     * <b>not</b> cause the server to start listening. To do so, create a new
+     * <code>Thread</code> wrapping this
+     * <code>NGServer</code> and start it.
      *
-     * @param addr the InetAddress to bind to
-     * @param port the port on which to listen
+     * @param listeningAddress the address at which to listen
      * @param sessionPoolSize the max number of idle sessions allowed by the
      * pool
+     * @param timeoutMillis timeout in millis to wait for a heartbeat from the client
+     * before disconnecting them
      */
-    private void init(InetAddress addr, int port, int sessionPoolSize) {
-        this.addr = addr;
-        this.port = port;
+    public NGServer(NGListeningAddress listeningAddress, int sessionPoolSize, int timeoutMillis) {
+        this.listeningAddress = listeningAddress;
 
-        this.aliasManager = new AliasManager();
+        aliasManager = new AliasManager();
         allNailStats = new java.util.HashMap();
         // allow a maximum of 10 idle threads.  probably too high a number
         // and definitely should be configurable in the future
         sessionPool = new NGSessionPool(this, sessionPoolSize);
+        heartbeatTimeoutMillis = timeoutMillis;
     }
 
     /**
@@ -266,7 +266,10 @@ public class NGServer implements Runnable {
      * @param nailClass the nail class that finished
      */
     void nailFinished(Class nailClass) {
-        NailStats stats = (NailStats) allNailStats.get(nailClass);
+        NailStats stats;
+        synchronized (allNailStats) {
+            stats = (NailStats) allNailStats.get(nailClass);
+        }
         stats.nailFinished();
     }
 
@@ -313,11 +316,8 @@ public class NGServer implements Runnable {
      * by your nails.
      */
     public void shutdown(boolean exitVM) {
-        synchronized (this) {
-            if (shutdown) {
-                return;
-            }
-            shutdown = true;
+        if (shutdown.getAndSet(true)) {
+            return;
         }
 
         try {
@@ -372,7 +372,7 @@ public class NGServer implements Runnable {
      * @return true iff the server is currently running.
      */
     public boolean isRunning() {
-        return (running);
+        return running.get();
     }
 
     /**
@@ -381,7 +381,7 @@ public class NGServer implements Runnable {
      * @return the port on which this server is (or will be) listening.
      */
     public int getPort() {
-        return ((serversocket == null) ? port : serversocket.getLocalPort());
+        return ((serversocket == null) ? listeningAddress.getInetPort() : serversocket.getLocalPort());
     }
 
     /**
@@ -389,7 +389,7 @@ public class NGServer implements Runnable {
      * them.
      */
     public void run() {
-        running = true;
+        running.set(true);
         NGSession sessionOnDeck = null;
 
         originalSecurityManager = System.getSecurityManager();
@@ -407,13 +407,49 @@ public class NGServer implements Runnable {
         }
 
         try {
-            if (addr == null) {
-                serversocket = new ServerSocket(port);
+            if (listeningAddress.isInetAddress()) {
+                if (listeningAddress.getInetAddress() == null) {
+                    serversocket = new ServerSocket(listeningAddress.getInetPort());
+                } else {
+                    serversocket = new ServerSocket(listeningAddress.getInetPort(), 0, listeningAddress.getInetAddress());
+                }
             } else {
-                serversocket = new ServerSocket(port, 0, addr);
+                if (Platform.isWindows()) {
+                    boolean requireStrictLength = true;
+                    serversocket = new NGWin32NamedPipeServerSocket(listeningAddress.getLocalAddress(), requireStrictLength);
+                } else {
+                    serversocket = new NGUnixDomainServerSocket(listeningAddress.getLocalAddress());
+                }
             }
 
-            while (!shutdown) {
+            String portDescription;
+            if (listeningAddress.isInetAddress() && listeningAddress.getInetPort() == 0) {
+                // if the port is 0, it will be automatically determined.
+                // add this little wait so the ServerSocket can fully
+                // initialize and we can see what port it chose.
+                int runningPort = getPort();
+                while (runningPort == 0) {
+                    try {
+                        Thread.sleep(50);
+                    } catch (Throwable toIgnore) {
+                    }
+                    runningPort = getPort();
+                }
+                portDescription = ", port " + runningPort;
+            } else {
+                portDescription = "";
+            }
+
+            // Only after this point nailgun server is ready to accept connections on all platforms.
+            // test_ng.py on *nix relies on reading this line from stdout to start connecting to server.
+            out.println("NGServer "
+                    + NGConstants.VERSION
+                    + " started on "
+                    + listeningAddress.toString()
+                    + portDescription
+                    + ".");
+
+            while (!shutdown.get()) {
                 sessionOnDeck = sessionPool.take();
                 Socket socket = serversocket.accept();
                 sessionOnDeck.run(socket);
@@ -423,14 +459,14 @@ public class NGServer implements Runnable {
             // if shutdown is called while the accept() method is blocking,
             // an exception will be thrown that we don't care about.  filter
             // those out.
-            if (!shutdown) {
+            if (!shutdown.get()) {
                 t.printStackTrace();
             }
         }
         if (sessionOnDeck != null) {
             sessionOnDeck.shutdown();
         }
-        running = false;
+        running.set(false);
     }
 
     private static void usage() {
@@ -438,6 +474,7 @@ public class NGServer implements Runnable {
         System.err.println("   or: java com.martiansoftware.nailgun.NGServer port");
         System.err.println("   or: java com.martiansoftware.nailgun.NGServer IPAddress");
         System.err.println("   or: java com.martiansoftware.nailgun.NGServer IPAddress:port");
+        System.err.println("   or: java com.martiansoftware.nailgun.NGServer IPAddress:port timeout");
     }
 
     /**
@@ -453,19 +490,24 @@ public class NGServer implements Runnable {
      */
     public static void main(String[] args) throws NumberFormatException, UnknownHostException {
 
-        if (args.length > 1) {
+        if (args.length > 2) {
             usage();
             return;
         }
 
         // null server address means bind to everything local
-        InetAddress serverAddress = null;
-        int port = NGConstants.DEFAULT_PORT;
+        NGListeningAddress listeningAddress;
+        int timeoutMillis = NGConstants.HEARTBEAT_TIMEOUT_MILLIS;
 
-        // parse the sole command line parameter, which
+        // parse the command line parameters, which
         // may be an inetaddress to bind to, a port number,
-        // or an inetaddress followed by a port, separated
-        // by a colon
+        // an inetaddress followed by a port, separated
+        // by a colon, or the string "local:/path/to/socket"
+        // for a Unix domain socket or Windows named pipe.
+        // if a second parameter is provided it
+        // is interpreted as the number of milliseconds to
+        // wait between heartbeats before considering the
+        // client to have disconnected.
         if (args.length != 0) {
             String[] argParts = args[0].split(":");
             String addrPart = null;
@@ -478,42 +520,36 @@ public class NGServer implements Runnable {
             } else {
                 portPart = argParts[0];
             }
-            if (addrPart != null) {
-                serverAddress = InetAddress.getByName(addrPart);
+            if ("local".equals(addrPart) && portPart != null) {
+                // Treat the port part as a path to a local Unix domain socket
+                // or Windows named pipe.
+                listeningAddress = new NGListeningAddress(portPart);
+            } else if (addrPart != null && portPart != null) {
+                listeningAddress = new NGListeningAddress(
+                    InetAddress.getByName(addrPart), Integer.parseInt(portPart));
+            } else if (addrPart != null && portPart == null) {
+                listeningAddress = new NGListeningAddress(
+                    InetAddress.getByName(addrPart), NGConstants.DEFAULT_PORT);
+            } else {
+                listeningAddress = new NGListeningAddress(null, Integer.parseInt(portPart));
             }
-            if (portPart != null) {
-                port = Integer.parseInt(portPart);
+            if (args.length == 2) {
+                timeoutMillis = Integer.parseInt(args[1]);
             }
+        } else {
+            listeningAddress = new NGListeningAddress(null, NGConstants.DEFAULT_PORT);
         }
 
-        NGServer server = new NGServer(serverAddress, port, DEFAULT_SESSIONPOOLSIZE);
+        NGServer server = new NGServer(listeningAddress, DEFAULT_SESSIONPOOLSIZE, timeoutMillis);
         Thread t = new Thread(server);
-        t.setName("NGServer(" + serverAddress + ", " + port + ")");
+        t.setName("NGServer(" + listeningAddress.toString() + ")");
         t.start();
 
         Runtime.getRuntime().addShutdownHook(new NGServerShutdowner(server));
+    }
 
-        // if the port is 0, it will be automatically determined.
-        // add this little wait so the ServerSocket can fully
-        // initialize and we can see what port it chose.
-        int runningPort = server.getPort();
-        while (runningPort == 0) {
-            try {
-                Thread.sleep(50);
-            } catch (Throwable toIgnore) {
-            }
-            runningPort = server.getPort();
-        }
-
-        System.out.println("NGServer "
-                + NGConstants.VERSION
-                + " started on "
-                + ((serverAddress == null)
-                ? "all interfaces"
-                : serverAddress.getHostAddress())
-                + ", port "
-                + runningPort
-                + ".");
+    public int getHeartbeatTimeout() {
+        return heartbeatTimeoutMillis;
     }
 
     /**
